@@ -1,35 +1,5 @@
-/* ----------------------------------------------------------------------- *
- *
- *   Copyright 1996-2020 The NASM Authors - All Rights Reserved
- *   See the file AUTHORS included with the NASM distribution for
- *   the specific copyright holders.
- *
- *   Redistribution and use in source and binary forms, with or without
- *   modification, are permitted provided that the following
- *   conditions are met:
- *
- *   * Redistributions of source code must retain the above copyright
- *     notice, this list of conditions and the following disclaimer.
- *   * Redistributions in binary form must reproduce the above
- *     copyright notice, this list of conditions and the following
- *     disclaimer in the documentation and/or other materials provided
- *     with the distribution.
- *
- *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
- *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
- *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
- *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- *     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
- *     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- *     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- *     NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- *     HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- *     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
- *     OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
- *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- * ----------------------------------------------------------------------- */
+/* SPDX-License-Identifier: BSD-2-Clause */
+/* Copyright 1996-2020 The NASM Authors - All Rights Reserved */
 
 /*
  * listing.c    listing file generator for the Netwide Assembler
@@ -54,6 +24,7 @@ static const char xdigit[] = "0123456789ABCDEF";
 #define HEX(a,b) (*(a)=xdigit[((b)>>4)&15],(a)[1]=xdigit[(b)&15]);
 
 uint64_t list_options, active_list_options;
+bool user_nolist;
 
 static char listline[LIST_MAX_LEN];
 static bool listlinep;
@@ -71,6 +42,20 @@ static int listlevel, listlevel_e;
 
 static FILE *listfp;
 
+static inline char err_fill_char(errflags severity)
+{
+    severity &= ERR_MASK;
+
+    if (severity < ERR_NOTE)
+        return ' ';
+    else if (severity < ERR_WARNING)
+        return '-';
+    else if (severity < ERR_CRITICAL)
+        return '*';
+    else
+        return 'X';
+}
+
 static void list_emit(void)
 {
     int i;
@@ -100,12 +85,11 @@ static void list_emit(void)
     }
 
     if (list_errors) {
-        static const char fillchars[] = " --***XX";
-        char fillchar;
-
         strlist_for_each(e, list_errors) {
+            char fillchar;
+
             fprintf(listfp, "%6"PRId32"          ", listlineno);
-            fillchar = fillchars[e->pvt.u & ERR_MASK];
+            fillchar = err_fill_char(e->pvt.u);
             for (i = 0; i < LIST_HEXBIT; i++)
                 putc(fillchar, listfp);
 
@@ -161,6 +145,7 @@ static void list_init(const char *fname)
     list_errors = NULL;
     listlevel = 0;
     suppress = 0;
+    user_nolist = false;
 }
 
 static void list_out(int64_t offset, char *str)
@@ -211,7 +196,7 @@ static void list_output(const struct out_data *data)
 {
     char q[24];
     uint64_t size = data->size;
-    uint64_t offset = data->offset;
+    uint64_t offset = data->loc.offset;
     const uint8_t *p = data->data;
 
 
@@ -231,7 +216,7 @@ static void list_output(const struct out_data *data)
     {
 	if (size == 0) {
             if (!listdata[0])
-                listoffset = data->offset;
+                listoffset = data->loc.offset;
         } else if (p) {
             while (size--) {
                 HEX(q, *p);
