@@ -1,35 +1,5 @@
-/* ----------------------------------------------------------------------- *
- *
- *   Copyright 1996-2023 The NASM Authors - All Rights Reserved
- *   See the file AUTHORS included with the NASM distribution for
- *   the specific copyright holders.
- *
- *   Redistribution and use in source and binary forms, with or without
- *   modification, are permitted provided that the following
- *   conditions are met:
- *
- *   * Redistributions of source code must retain the above copyright
- *     notice, this list of conditions and the following disclaimer.
- *   * Redistributions in binary form must reproduce the above
- *     copyright notice, this list of conditions and the following
- *     disclaimer in the documentation and/or other materials provided
- *     with the distribution.
- *
- *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
- *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
- *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
- *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- *     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
- *     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- *     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- *     NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- *     HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- *     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
- *     OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
- *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- * ----------------------------------------------------------------------- */
+/* SPDX-License-Identifier: BSD-2-Clause */
+/* Copyright 1996-2025 The NASM Authors - All Rights Reserved */
 
 /*
  * nasmlib.h    header file for nasmlib.c
@@ -54,6 +24,67 @@ union intorptr {
 };
 typedef union intorptr intorptr;
 
+/*
+ * Handy macro to use as the base of shifts
+ */
+#define ONE ((uintmax_t)1)
+
+/*
+ * Handy macros for defining and accessing enums of bitfields;
+ * provides a set of standard-named parameters for each field.
+ *
+ * mk_field_mask() is indended to be used for non-contiguous fields.
+ */
+
+#define mk_field_mask(name,pos,width,basemask)              \
+    name ## _POS = (pos),                                   \
+    name ## _WIDTH = (width),                               \
+    name ## _BASEMASK = (basemask),                         \
+    name ## _MASK = name ## _BASEMASK << name ## _POS,      \
+    name = name ## _MASK
+
+#define mk_field(name,pos,width)                            \
+    mk_field_mask(name,pos,width,                           \
+                  (1 << name ## _WIDTH)-1)
+
+/*
+ * Cast a value to a suitable type to represent the encoded
+ * (post-shift) and extracted (pre-shift) values, respectively.
+ */
+#ifdef HAVE_TYPEOF
+# define fieldenc_cast(x,y) ((typeof(x ## _MASK))(y))
+# define fieldval_cast(x,y) ((typeof(x ## _BASEMASK))(y))
+#else
+/*
+ * Using int here matches the strict ISO C before C23 which only allows
+ * an enum to be in the range of "int". This also means that these macros
+ * are only usable for bitfields up to 32 bits wide, which is extremely
+ * unfortunate, but covers most of all the NASM use cases.
+ *
+ * The (int) should at least give a warning on overflow.
+ */
+# define fieldenc_cast(x,y) ((int)(y))
+# define fieldval_cast(x,y) (y)
+#endif
+
+/*
+ * Macros to get/set bitfield values (the set macro returns the
+ * updated value, it does not change the input.)
+ *
+ * The fieldenc() macro produces the masked and shifted value
+ * corresponding to a base value; it is equivalent to
+ * setfield(field,0,val).
+ */
+#define getfield(field,from)                            \
+    fieldval_cast(field,                                \
+                  (((from) >> field ## _POS) &          \
+                   field ## _BASEMASK))
+#define fieldval(field,val)                                \
+    (((val) & fieldenc_cast(field,field ## _BASEMASK))     \
+     << field ## _POS)
+#define setfield(field,from,val)                        \
+    (((from) & ~(field ## _MASK)) + fieldval(field,val))
+
 /*
  * Wrappers around malloc, realloc, free and a few more. nasm_malloc
  * will fatal-error and die rather than return NULL; nasm_realloc will
@@ -75,11 +106,11 @@ char * safe_alloc end_with_null nasm_strcatn(const char *one, ...);
  * nasm_[v]asprintf() are variants of the semi-standard [v]asprintf()
  * functions, except that we return the pointer instead of a count.
  * The size of the string (including the final NUL!) is available
- * by calling nasm_aprintf_size() afterwards.
+ * by calling nasm_last_string_size() afterwards.
  *
  * nasm_[v]axprintf() are similar, but allocates a user-defined amount
  * of storage before the string, and returns a pointer to the
- * allocated buffer. The value of nasm_aprintf_size() does *not* include
+ * allocated buffer. The value of nasm_last_string_size() does *not* include
  * this additional storage.
  */
 char * safe_alloc printf_func(1, 2) nasm_asprintf(const char *fmt, ...);
@@ -105,7 +136,7 @@ static inline size_t nasm_last_string_size(void)
     return _nasm_last_string_size;
 }
 
-/* Assert the argument is a pointer without evaluating it */
+/* Statically assert the argument is a pointer without evaluating it */
 #define nasm_assert_pointer(p) ((void)sizeof(*(p)))
 
 #define nasm_new(p) ((p) = nasm_zalloc(sizeof(*(p))))
@@ -181,12 +212,23 @@ void nasm_write(const void *, size_t, FILE *);
 /*
  * NASM assert failure
  */
-fatal_func nasm_assert_failed(const char *, int, const char *);
-#define nasm_assert(x)                                          \
-    do {                                                        \
-        nasm_try_static_assert(x);                              \
-        if (unlikely(!(x)))                                     \
-            nasm_assert_failed(__FILE__,__LINE__,#x);           \
+fatal_func nasm_assert_failed(const char *msg, const char *func,
+                              const char *file, int line);
+
+/* Plain assert, gives the source location as an error message */
+#define nasm_assert(x)                                                  \
+    do {                                                                \
+        nasm_try_static_assert(x);                                      \
+        if (unlikely(!(x)))                                             \
+            nasm_assert_failed(#x,NASM_FUNC,__FILE__,__LINE__);         \
+    } while (0)
+
+/* Assert with custom message */
+#define nasm_assert_msg(x,m)                    \
+    do {                                        \
+        nasm_try_static_assert(x);              \
+        if (unlikely(!(x)))                     \
+            nasm_panic("%s", m);                \
     } while (0)
 
 /* Utility function to generate a string for an invalid enum */
@@ -237,6 +279,11 @@ static inline unsigned int numvalue(unsigned char c)
  */
 int64_t readnum(const char *str, bool *error);
 
+/*
+ * Warn for the use of $ as a hexadecimal prefix
+ */
+void warn_dollar_hex(void);
+
 /*
  * Get the numeric base corresponding to a character
  */
@@ -276,6 +323,12 @@ int64_t readstrnum(char *str, int length, bool *warn);
 int numstr(char *buf, size_t buflen, uint64_t n,
            int digits, unsigned int base, bool ucase);
 
+extern const char * const nasmlib_digit_chars[2];
+static inline const char *nasm_digit_chars(bool ucase)
+{
+    return nasmlib_digit_chars[ucase];
+}
+
 /*
  * seg_alloc: allocate a hitherto unused segment number.
  */
@@ -293,6 +346,10 @@ const char *filename_set_extension(const char *inname, const char *extension);
  * the number of elements of a statically defined array.
  */
 #define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))
+#define ARRAY_END(arr)  (&(arr)[ARRAY_SIZE(arr)])
+#define array_for_each(var,arr) \
+    for ((var) = (arr); (var) < ARRAY_END(arr); (var)++)
+
 
 /*
  * List handling
@@ -353,14 +410,14 @@ void fwriteaddr(uint64_t data, int size, FILE * fp);
  *
  * bsi() is case sensitive, bsii() is case insensitive.
  */
-int bsi(const char *string, const char **array, int size);
-int bsii(const char *string, const char **array, int size);
+int pure_func bsi(const char *string, const char **array, int size);
+int pure_func bsii(const char *string, const char **array, int size);
 
 /*
  * Convenient string processing helper routines
  */
-char *nasm_skip_spaces(const char *p);
-char *nasm_skip_word(const char *p);
+char * pure_func nasm_skip_spaces(const char *p);
+char * pure_func nasm_skip_word(const char *p);
 char *nasm_zap_spaces_fwd(char *p);
 char *nasm_zap_spaces_rev(char *p);
 char *nasm_trim_spaces(char *p);
@@ -381,7 +438,12 @@ char * safe_alloc nasm_dirname(const char *path);
 char * safe_alloc nasm_basename(const char *path);
 char * safe_alloc nasm_catfile(const char *dir, const char *path);
 
-const char * pure_func prefix_name(int);
+/*
+ * Various tokens to readable strings, with limit checking
+ */
+const char * const_func register_name(int);
+const char * const_func prefix_name(int);
+bool const_func is_hint_nop(uint64_t);
 
 /*
  * Wrappers around fopen()... for future change to a dedicated structure
@@ -434,59 +496,71 @@ off_t nasm_file_size_by_path(const char *pathname);
 bool nasm_file_time(time_t *t, const char *pathname);
 void fwritezero(off_t bytes, FILE *fp);
 
-static inline bool const_func overflow_general(int64_t value, int bytes)
+/* Sign-extend a value to an arbitrary number of bits */
+static inline int64_t const_func sext(int64_t value, unsigned int bits)
 {
-    int sbit;
-    int64_t vmax, vmin;
-
-    if (bytes >= 8)
-        return false;
-
-    sbit = (bytes << 3) - 1;
-    vmax =  ((int64_t)2 << sbit) - 1;
-    vmin = -((int64_t)2 << sbit);
-
-    return value < vmin || value > vmax;
-}
-
-static inline bool const_func overflow_signed(int64_t value, int bytes)
-{
-    int sbit;
-    int64_t vmax, vmin;
-
-    if (bytes >= 8)
-        return false;
-
-    sbit = (bytes << 3) - 1;
-    vmax =  ((int64_t)1 << sbit) - 1;
-    vmin = -((int64_t)1 << sbit);
-
-    return value < vmin || value > vmax;
-}
-
-static inline bool const_func overflow_unsigned(int64_t value, int bytes)
-{
-    int sbit;
-    int64_t vmax, vmin;
-
-    if (bytes >= 8)
-        return false;
-
-    sbit = (bytes << 3) - 1;
-    vmax = ((int64_t)2 << sbit) - 1;
-    vmin = 0;
-
-    return value < vmin || value > vmax;
-}
-
-static inline int64_t const_func signed_bits(int64_t value, int bits)
-{
-    if (bits < 64) {
-        value &= ((int64_t)1 << bits) - 1;
-        if (value & (int64_t)1 << (bits - 1))
-            value |= (int64_t)((uint64_t)-1 << bits);
+    if (is_constant(bits)) {
+        switch (bits) {
+        case 8:
+            return (int8_t)value;
+        case 16:
+            return (int16_t)value;
+        case 32:
+            return (int32_t)value;
+        default:
+            break;
+        }
     }
-    return value;
+
+    if (bits >= 64)
+        return value;
+
+    /* sext(foo,0) == sext(foo,1) */
+    bits = bits ? 64-bits : 63;
+
+    return value << bits >> bits;
+}
+
+/* Zero-extend a value to an arbitrary number of bits */
+static inline uint64_t const_func zext(uint64_t value, unsigned int bits)
+{
+    if (is_constant(bits)) {
+        switch (bits) {
+        case 8:
+            return (uint8_t)value;
+        case 16:
+            return (uint16_t)value;
+        case 32:
+            return (uint32_t)value;
+        default:
+            break;
+        }
+    }
+
+    if (bits >= 64)
+        return value;
+
+    if (!bits)
+        return 0;
+
+    bits = 64-bits;
+    return value << bits >> bits;
+}
+
+static inline bool const_func overflow_signed(int64_t value, unsigned int bytes)
+{
+    return sext(value, bytes << 3) != value;
+}
+
+static inline bool const_func overflow_unsigned(uint64_t value, unsigned int bytes)
+{
+    return zext(value, bytes << 3) != value;
+}
+
+/* This is very conservative, but otherwise things like ~0x80 break */
+static inline bool const_func overflow_general(int64_t value, unsigned int bytes)
+{
+    return overflow_unsigned(value, bytes) && overflow_unsigned(-value, bytes);
 }
 
 /* check if value is power of 2 */
