@@ -1,35 +1,5 @@
-/* ----------------------------------------------------------------------- *
- *
- *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
- *   See the file AUTHORS included with the NASM distribution for
- *   the specific copyright holders.
- *
- *   Redistribution and use in source and binary forms, with or without
- *   modification, are permitted provided that the following
- *   conditions are met:
- *
- *   * Redistributions of source code must retain the above copyright
- *     notice, this list of conditions and the following disclaimer.
- *   * Redistributions in binary form must reproduce the above
- *     copyright notice, this list of conditions and the following
- *     disclaimer in the documentation and/or other materials provided
- *     with the distribution.
- *
- *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
- *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
- *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
- *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- *     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
- *     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- *     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- *     NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- *     HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- *     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
- *     OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
- *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- * ----------------------------------------------------------------------- */
+/* SPDX-License-Identifier: BSD-2-Clause */
+/* Copyright 1996-2025 The NASM Authors - All Rights Reserved */
 
 /*
  * nasmlib.c	library routines for the Netwide Assembler
@@ -41,10 +11,24 @@
 
 #include "nasmlib.h"
 #include "error.h"
-#include "nasm.h"               /* For globalbits */
+#include "nasm.h"               /* For globl.dollarhex */
 
 #define lib_isnumchar(c)    (nasm_isalnum(c) || (c) == '$' || (c) == '_')
 
+void warn_dollar_hex(void)
+{
+    /*!
+     *!number-deprecated-hex [on] $ prefix for hexadecimal is deprecated
+     *!    warns that the \c{$} prefix for hexadecimal numbers is
+     *!    deprecated, due to the syntactic conflict with \c{$} used
+     *!    as a symbol escape prefix. This syntax may be disabled by
+     *!    default in a future version of NASM. Replace \c{$} with \c{0x}
+     *!    to ensure compatibility with future versions.
+     */
+    nasm_warn(WARN_NUMBER_DEPRECATED_HEX,
+              "$ prefix for hexadecimal is deprecated");
+}
+
 int64_t readnum(const char *str, bool *error)
 {
     const char *r = str, *q;
@@ -91,10 +75,12 @@ int64_t readnum(const char *str, bool *error)
     pradix = sradix = 0;
     plen = slen = 0;
 
-    if (len > 2 && *r == '0' && (pradix = radix_letter(r[1])) != 0)
+    if (len > 2 && *r == '0' && (pradix = radix_letter(r[1])) != 0) {
 	plen = 2;
-    else if (len > 1 && *r == '$')
+    } else if (len > 1 && *r == '$' && globl.dollarhex) {
+        /* Warning here would probably duplicate warnings */
 	pradix = 16, plen = 1;
+    }
 
     if (len > 1 && (sradix = radix_letter(q[-1])) != 0)
 	slen = 1;
