@@ -1,35 +1,5 @@
-/* ----------------------------------------------------------------------- *
- *
- *   Copyright 1996-2017 The NASM Authors - All Rights Reserved
- *   See the file AUTHORS included with the NASM distribution for
- *   the specific copyright holders.
- *
- *   Redistribution and use in source and binary forms, with or without
- *   modification, are permitted provided that the following
- *   conditions are met:
- *
- *   * Redistributions of source code must retain the above copyright
- *     notice, this list of conditions and the following disclaimer.
- *   * Redistributions in binary form must reproduce the above
- *     copyright notice, this list of conditions and the following
- *     disclaimer in the documentation and/or other materials provided
- *     with the distribution.
- *
- *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
- *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
- *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
- *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- *     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
- *     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- *     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- *     NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- *     HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- *     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
- *     OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
- *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- * ----------------------------------------------------------------------- */
+/* SPDX-License-Identifier: BSD-2-Clause */
+/* Copyright 1996-2025 The NASM Authors - All Rights Reserved */
 
 /*
  * outobj.c	output routines for the Netwide Assembler to produce
@@ -43,6 +13,7 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "asmutil.h"
 #include "error.h"
 #include "stdscan.h"
 #include "eval.h"
@@ -580,9 +551,10 @@ static struct Segment {
     char *name;
     int32_t index;                 /* the NASM segment id */
     int32_t obj_index;             /* the OBJ-file segment index */
-    struct Group *grp;          /* the group it beint32_ts to */
+    struct Group *grp;             /* the group it beint32_ts to */
     uint32_t currentpos;
-    int32_t align;                 /* can be SEG_ABS + absolute addr */
+    int32_t align;                /* can be SEG_ABS + absolute addr */
+    uint64_t origalign;           /* originally requested alignment */
     int64_t pass_last_seen;
     struct Public *pubhead, **pubtail, *lochead, **loctail;
     char *segclass, *overlay;   /* `class' is a C++ keyword :-) */
@@ -638,6 +610,9 @@ static const struct dfmt borland_debug_form;
 /* The current segment */
 static struct Segment *current_seg;
 
+/* Name for segment to use if no segment directive is defined */
+static char DEFAULT_SEG[] = "__NASMDEFSEG";
+
 static int32_t obj_segment(char *, int *);
 static void obj_write_file(void);
 static enum directive_result obj_directive(enum directive, char *);
@@ -777,9 +752,8 @@ static void obj_deflabel(char *name, int32_t segment,
     int i;
     bool used_special = false;   /* have we used the special text? */
 
-    if (debug_level(2))
-        nasm_debug(" obj_deflabel: %s, seg=%"PRIx32", off=%"PRIx64", is_global=%d, %s\n",
-                   name, segment, offset, is_global, special);
+    nasm_debug(2, " obj_deflabel: %s, seg=%"PRIx32", off=%"PRIx64", is_global=%d, %s\n",
+               name, segment, offset, is_global, special);
 
     /*
      * If it's a special-retry from pass two, discard it.
@@ -840,7 +814,7 @@ static void obj_deflabel(char *name, int32_t segment,
      */
     if (!any_segs && segment == first_seg) {
         int tempint = 0;
-        if (segment != obj_segment("__NASMDEFSEG", &tempint))
+        if (segment != obj_segment(DEFAULT_SEG, &tempint))
             nasm_panic("strange segment conditions in OBJ driver");
     }
 
@@ -964,8 +938,7 @@ static void obj_deflabel(char *name, int32_t segment,
                 expr *e;
                 struct tokenval tokval;
 
-                stdscan_reset();
-                stdscan_set(special);
+                stdscan_reset(special);
                 tokval.t_type = TOKEN_INVALID;
                 e = evaluate(stdscan, NULL, &tokval, NULL, 1, NULL);
                 if (e) {
@@ -975,7 +948,7 @@ static void obj_deflabel(char *name, int32_t segment,
                     else
                         ext->commonelem = reloc_value(e);
                 }
-                special = stdscan_get();
+                special = stdscan_tell();
             } else {
                 nasm_nonfatal("`%s': element-size specifications only"
                               " apply to common variables", ext->name);
@@ -1017,10 +990,9 @@ static void obj_write_fixup(ObjRecord * orp, int bytes,
                             int segrel, int32_t seg, int32_t wrt,
                             struct Segment *segto);
 
-static void obj_out(int32_t segto, const void *data,
-		    enum out_type type, uint64_t size,
-                    int32_t segment, int32_t wrt)
+static void obj_out(const struct out_data *out)
 {
+    OUT_LEGACY(out,segto,data,type,size,segment,wrt);
     const uint8_t *ucdata;
     int32_t ldata;
     struct Segment *seg;
@@ -1032,7 +1004,7 @@ static void obj_out(int32_t segto, const void *data,
      */
     if (!any_segs) {
         int tempint = 0;
-        if (segto != obj_segment("__NASMDEFSEG", &tempint))
+        if (segto != obj_segment(DEFAULT_SEG, &tempint))
             nasm_panic("strange segment conditions in OBJ driver");
     }
 
@@ -1326,6 +1298,40 @@ static void obj_write_fixup(ObjRecord * orp, int bytes,
     obj_commit(forp);
 }
 
+static uint32_t check_segment_alignment(const uint64_t origalign)
+{
+    /* Supported alignment values */
+    const uint32_t alignments = 1 | 2 | 4 | 16 | 256 | 4096;
+    uint32_t align = origalign;
+
+    if (origalign &&
+        (!is_power2(origalign) || origalign > alignments)) {
+        nasm_nonfatal("invalid alignment value %"PRIu64, origalign);
+        return 1;
+    }
+
+    if (align == 0)
+        align = 1;
+
+    if (align & alignments)
+        return align;           /* All good! */
+
+    while (!(align & alignments))
+        align <<= 1;
+
+    /*!
+     *!section-alignment-rounded [on] section alignment rounded up
+     *!  warn if a section alignment is specified which is
+     *!  not supported by the underlying object format, but
+     *!  can be rounded up to a supported value.
+     */
+    nasm_warn(WARN_SECTION_ALIGNMENT_ROUNDED,
+              "alignment of %"PRIu64" not supported, using %"PRIu32,
+              origalign, align);
+
+    return align;
+}
+
 static int32_t obj_segment(char *name, int *bits)
 {
     /*
@@ -1335,8 +1341,7 @@ static int32_t obj_segment(char *name, int *bits)
      * using the pointer it gets passed. That way we save memory,
      * by sponging off the label manager.
      */
-    if (debug_level(3))
-        nasm_debug(" obj_segment: < %s >, *bits=%d\n", name, *bits);
+    nasm_debug(3, " obj_segment: < %s >, *bits=%d\n", name, *bits);
 
     if (!name) {
         *bits = 16;
@@ -1346,34 +1351,19 @@ static int32_t obj_segment(char *name, int *bits)
         struct Segment *seg;
         struct Group *grp;
         struct External **extp;
-        int obj_idx, i, attrs;
+        int obj_idx, i;
 	bool rn_error;
         char *p;
 
         /*
          * Look for segment attributes.
          */
-        attrs = 0;
         while (*name == '.')
             name++;             /* hack, but a documented one */
-        p = name;
-        while (*p && !nasm_isspace(*p))
-            p++;
+        p = nasm_skip_word(name);
         if (*p) {
             *p++ = '\0';
-            while (*p && nasm_isspace(*p))
-                *p++ = '\0';
-        }
-        while (*p) {
-            while (*p && !nasm_isspace(*p))
-                p++;
-            if (*p) {
-                *p++ = '\0';
-                while (*p && nasm_isspace(*p))
-                    *p++ = '\0';
-            }
-
-            attrs++;
+            p = nasm_skip_spaces(p);
         }
 
         for (seg = seghead, obj_idx = 1; ; seg = seg->next, obj_idx++) {
@@ -1381,9 +1371,16 @@ static int32_t obj_segment(char *name, int *bits)
                 break;
 
             if (!strcmp(seg->name, name)) {
-                if (attrs > 0 && seg->pass_last_seen == pass_count())
-                    nasm_warn(WARN_OTHER, "segment attributes specified on"
-                              " redeclaration of segment: ignoring");
+                if (seg->pass_last_seen == pass_count()) {
+                    if (*p)
+                        nasm_warn(WARN_OTHER, "segment attributes specified on"
+                                  " redeclaration of segment: ignoring");
+                } else {
+                    /* Reissue alignment warning on this pass if necessary */
+                    if (seg->align < SEG_ABS)
+                        check_segment_alignment(seg->origalign);
+                }
+
                 if (seg->use32)
                     *bits = 32;
                 else
@@ -1403,9 +1400,9 @@ static int32_t obj_segment(char *name, int *bits)
         any_segs = true;
         seg->name = nasm_strdup(name);
         seg->currentpos = 0;
-        seg->align = 1;         /* default */
-        seg->use32 = false;     /* default */
-        seg->combine = CMB_PUBLIC;      /* default */
+        seg->align = seg->origalign = 1; /* default */
+        seg->use32 = false;              /* default */
+        seg->combine = CMB_PUBLIC;       /* default */
         seg->segclass = seg->overlay = NULL;
         seg->pubhead = NULL;
         seg->pubtail = &seg->pubhead;
@@ -1420,28 +1417,30 @@ static int32_t obj_segment(char *name, int *bits)
         /*
          * Process the segment attributes.
          */
-        p = name;
-        while (attrs--) {
-            p += strlen(p);
-            while (!*p)
-                p++;
+        while (*p) {
+            const char *q = p;
+            p = nasm_skip_word(p);
+            if (*p) {
+                *p++ = '\0';
+                p = nasm_skip_spaces(p);
+            }
 
             /*
              * `p' contains a segment attribute.
              */
-            if (!nasm_stricmp(p, "private"))
+            if (!nasm_stricmp(q, "private"))
                 seg->combine = CMB_PRIVATE;
-            else if (!nasm_stricmp(p, "public"))
+            else if (!nasm_stricmp(q, "public"))
                 seg->combine = CMB_PUBLIC;
-            else if (!nasm_stricmp(p, "common"))
+            else if (!nasm_stricmp(q, "common"))
                 seg->combine = CMB_COMMON;
-            else if (!nasm_stricmp(p, "stack"))
+            else if (!nasm_stricmp(q, "stack"))
                 seg->combine = CMB_STACK;
-            else if (!nasm_stricmp(p, "use16"))
+            else if (!nasm_stricmp(q, "use16"))
                 seg->use32 = false;
-            else if (!nasm_stricmp(p, "use32"))
+            else if (!nasm_stricmp(q, "use32"))
                 seg->use32 = true;
-            else if (!nasm_stricmp(p, "flat")) {
+            else if (!nasm_stricmp(q, "flat")) {
                 /*
                  * This segment is an OS/2 FLAT segment. That means
                  * that its default group is group FLAT, even if
@@ -1466,66 +1465,52 @@ static int32_t obj_segment(char *name, int *bits)
                         nasm_panic("failure to define FLAT?!");
                 }
                 seg->grp = grp;
-            } else if (!nasm_strnicmp(p, "class=", 6))
-                seg->segclass = nasm_strdup(p + 6);
-            else if (!nasm_strnicmp(p, "overlay=", 8))
-                seg->overlay = nasm_strdup(p + 8);
-            else if (!nasm_strnicmp(p, "align=", 6)) {
-                seg->align = readnum(p + 6, &rn_error);
+            } else if (!nasm_strnicmp(q, "class=", 6))
+                seg->segclass = nasm_strdup(q + 6);
+            else if (!nasm_strnicmp(q, "overlay=", 8))
+                seg->overlay = nasm_strdup(q + 8);
+            else if (!nasm_strnicmp(q, "align=", 6)) {
+                uint64_t n = readnum(q + 6, &rn_error);
                 if (rn_error) {
-                    seg->align = 1;
                     nasm_nonfatal("segment alignment should be numeric");
+                    n = 1;
                 }
-                switch (seg->align) {
-                case 1:        /* BYTE */
-                case 2:        /* WORD */
-                case 4:        /* DWORD */
-                case 16:       /* PARA */
-                case 256:      /* PAGE */
-                case 4096:     /* PharLap extension */
-                    break;
-                case 8:
-                    nasm_warn(WARN_OTHER, "OBJ format does not support alignment"
-                              " of 8: rounding up to 16");
-                    seg->align = 16;
-                    break;
-                case 32:
-                case 64:
-                case 128:
-                    nasm_warn(WARN_OTHER, "OBJ format does not support alignment"
-                              " of %d: rounding up to 256", seg->align);
-                    seg->align = 256;
-                    break;
-                case 512:
-                case 1024:
-                case 2048:
-                    nasm_warn(WARN_OTHER, "OBJ format does not support alignment"
-                              " of %d: rounding up to 4096", seg->align);
-                    seg->align = 4096;
-                    break;
-                default:
-                    nasm_nonfatal("invalid alignment value %d",
-                                  seg->align);
-                    seg->align = 1;
-                    break;
-                }
-            } else if (!nasm_strnicmp(p, "absolute=", 9)) {
-                seg->align = SEG_ABS + readnum(p + 9, &rn_error);
-                if (rn_error)
+                seg->origalign = n;
+                seg->align = check_segment_alignment(n);
+            } else if (!nasm_strnicmp(q, "absolute=", 9)) {
+                uint64_t n = readnum(q + 9, &rn_error);
+                if (rn_error) {
                     nasm_nonfatal("argument to `absolute' segment"
                                   " attribute should be numeric");
+                    n = 0;
+                } else if (n >= SEG_ABS) {
+                    /* Probably could even be max 64K? */
+                    nasm_nonfatal("unsupported `absolute' segment number %#"PRIx64, n);
+                    n = 0;
+                }
+                seg->align = SEG_ABS + n;
             }
         }
 
+        if (!seg->use32 && seg->grp && !strcmp(seg->grp->name, "FLAT"))
+           nasm_panic("wrong combination of USE16(16-bit segment) and FLAT");
+
         /* We need to know whenever we have at least one 32-bit segment */
         obj_use32 |= seg->use32;
 
-        obj_seg_needs_update = seg;
-        if (seg->align >= SEG_ABS)
-            define_label(name, NO_SEG, seg->align - SEG_ABS, false);
-        else
-            define_label(name, seg->index + 1, 0L, false);
-        obj_seg_needs_update = NULL;
+        /*
+         * Trying to create a symbol on the last pass will end in tears.
+         * This can happen if there is no SEGMENT directive, and there
+         * are no labels.
+         */
+        if (!pass_final()) {
+            obj_seg_needs_update = seg;
+            if (seg->align >= SEG_ABS)
+                define_label(name, NO_SEG, seg->align - SEG_ABS, false);
+            else
+                define_label(name, seg->index + 1, 0L, false);
+            obj_seg_needs_update = NULL;
+        }
 
         /*
          * See if this segment is defined in any groups.
@@ -1536,7 +1521,12 @@ static int32_t obj_segment(char *name, int *bits)
                     nasm_free(grp->segs[i].name);
                     grp->segs[i] = grp->segs[grp->nindices];
                     grp->segs[grp->nindices++].index = seg->obj_index;
-                    if (seg->grp)
+                    /*
+                     * The group FLAT is a pseudo group. Therefore, it is
+                     * allowed to redefine a segment in the group FLAT as
+                     * other group.
+                     */
+                    if (seg->grp && strcmp(seg->grp->name, "FLAT"))
                         nasm_warn(WARN_OTHER, "segment `%s' is already part of"
                                   " a group: first one takes precedence",
                                   seg->name);
@@ -1577,9 +1567,9 @@ obj_directive(enum directive directive, char *value)
 {
     switch (directive) {
     case D_GROUP:
-    {
-        char *p, *q, *v;
-        if (pass_first()) {     /* XXX */
+        /* Is pass_first() really correct? */
+        if (value && pass_first()) {
+            char *p, *q, *v;
             struct Group *grp;
             struct Segment *seg;
             struct External **extp;
@@ -1651,7 +1641,12 @@ obj_directive(enum directive directive, char *value)
                      */
                     grp->segs[grp->nentries++] = grp->segs[grp->nindices];
                     grp->segs[grp->nindices++].index = seg->obj_index;
-                    if (seg->grp)
+                    /*
+                     * The group FLAT is a pseudo group. Therefore, it is
+                     * allowed to redefine a segment in the group FLAT as
+                     * other group.
+                     */
+                    if (seg->grp && strcmp(seg->grp->name, "FLAT"))
                         nasm_warn(WARN_OTHER, "segment `%s' is already part of"
                                   " a group: first one takes precedence",
                                   seg->name);
@@ -1684,95 +1679,18 @@ obj_directive(enum directive directive, char *value)
             }
         }
         return DIRR_OK;
-    }
+
     case D_UPPERCASE:
-        obj_uppercase = true;
+        if (value)
+            get_boolean_option(value, &obj_uppercase);
         return DIRR_OK;
 
     case D_IMPORT:
-    {
-        char *q, *extname, *libname, *impname;
+        /* Is pass_first() really correct? */
+        if (value && pass_first()) {
+            char *q, *extname, *libname, *impname;
 
-        if (!pass_first())      /* XXX */
-            return DIRR_OK;
-        extname = q = value;
-        while (*q && !nasm_isspace(*q))
-            q++;
-        if (nasm_isspace(*q)) {
-            *q++ = '\0';
-            while (*q && nasm_isspace(*q))
-                q++;
-        }
-
-        libname = q;
-        while (*q && !nasm_isspace(*q))
-            q++;
-        if (nasm_isspace(*q)) {
-            *q++ = '\0';
-            while (*q && nasm_isspace(*q))
-                q++;
-        }
-
-        impname = q;
-
-        if (!*extname || !*libname)
-            nasm_nonfatal("`import' directive requires symbol name"
-                          " and library name");
-        else {
-            struct ImpDef *imp;
-            bool err = false;
-
-            imp = *imptail = nasm_malloc(sizeof(struct ImpDef));
-            imptail = &imp->next;
-            imp->next = NULL;
-            imp->extname = nasm_strdup(extname);
-            imp->libname = nasm_strdup(libname);
-            imp->impindex = readnum(impname, &err);
-            if (!*impname || err)
-                imp->impname = nasm_strdup(impname);
-            else
-                imp->impname = NULL;
-        }
-
-        return DIRR_OK;
-    }
-    case D_EXPORT:
-    {
-        char *q, *extname, *intname, *v;
-        struct ExpDef *export;
-        int flags = 0;
-        unsigned int ordinal = 0;
-
-        if (!pass_first())
-            return DIRR_OK;     /* ignore in pass two */
-        intname = q = value;
-        while (*q && !nasm_isspace(*q))
-            q++;
-        if (nasm_isspace(*q)) {
-            *q++ = '\0';
-            while (*q && nasm_isspace(*q))
-                q++;
-        }
-
-        extname = q;
-        while (*q && !nasm_isspace(*q))
-            q++;
-        if (nasm_isspace(*q)) {
-            *q++ = '\0';
-            while (*q && nasm_isspace(*q))
-                q++;
-        }
-
-        if (!*intname) {
-            nasm_nonfatal("`export' directive requires export name");
-            return DIRR_OK;
-        }
-        if (!*extname) {
-            extname = intname;
-            intname = "";
-        }
-        while (*q) {
-            v = q;
+            extname = q = value;
             while (*q && !nasm_isspace(*q))
                 q++;
             if (nasm_isspace(*q)) {
@@ -1780,38 +1698,114 @@ obj_directive(enum directive directive, char *value)
                 while (*q && nasm_isspace(*q))
                     q++;
             }
-            if (!nasm_stricmp(v, "resident"))
-                flags |= EXPDEF_FLAG_RESIDENT;
-            else if (!nasm_stricmp(v, "nodata"))
-                flags |= EXPDEF_FLAG_NODATA;
-            else if (!nasm_strnicmp(v, "parm=", 5)) {
+
+            libname = q;
+            while (*q && !nasm_isspace(*q))
+                q++;
+            if (nasm_isspace(*q)) {
+                *q++ = '\0';
+                while (*q && nasm_isspace(*q))
+                    q++;
+            }
+
+            impname = q;
+
+            if (!*extname || !*libname)
+                nasm_nonfatal("`import' directive requires symbol name"
+                              " and library name");
+            else {
+                struct ImpDef *imp;
                 bool err = false;
-                flags |= EXPDEF_MASK_PARMCNT & readnum(v + 5, &err);
-                if (err) {
-                    nasm_nonfatal("value `%s' for `parm' is non-numeric", v + 5);
-                    return DIRR_ERROR;
-                }
-            } else {
-                bool err = false;
-                ordinal = readnum(v, &err);
-                if (err) {
-                    nasm_nonfatal("unrecognised export qualifier `%s'", v);
-                    return DIRR_ERROR;
-                }
-                flags |= EXPDEF_FLAG_ORDINAL;
+
+                imp = *imptail = nasm_malloc(sizeof(struct ImpDef));
+                imptail = &imp->next;
+                imp->next = NULL;
+                imp->extname = nasm_strdup(extname);
+                imp->libname = nasm_strdup(libname);
+                imp->impindex = readnum(impname, &err);
+                if (!*impname || err)
+                    imp->impname = nasm_strdup(impname);
+                else
+                    imp->impname = NULL;
             }
         }
-
-        export = *exptail = nasm_malloc(sizeof(struct ExpDef));
-        exptail = &export->next;
-        export->next = NULL;
-        export->extname = nasm_strdup(extname);
-        export->intname = nasm_strdup(intname);
-        export->ordinal = ordinal;
-        export->flags = flags;
-
         return DIRR_OK;
-    }
+
+    case D_EXPORT:
+        /* Is pass_first() really correct? */
+        if (value && pass_first()) {
+            char *q, *extname, *intname, *v;
+            struct ExpDef *export;
+            int flags = 0;
+            unsigned int ordinal = 0;
+
+            intname = q = value;
+            while (*q && !nasm_isspace(*q))
+                q++;
+            if (nasm_isspace(*q)) {
+                *q++ = '\0';
+                while (*q && nasm_isspace(*q))
+                    q++;
+            }
+
+            extname = q;
+            while (*q && !nasm_isspace(*q))
+                q++;
+            if (nasm_isspace(*q)) {
+                *q++ = '\0';
+                while (*q && nasm_isspace(*q))
+                    q++;
+            }
+
+            if (!*intname) {
+                nasm_nonfatal("`export' directive requires export name");
+                return DIRR_OK;
+            }
+            if (!*extname) {
+                extname = intname;
+                intname = "";
+            }
+            while (*q) {
+                v = q;
+                while (*q && !nasm_isspace(*q))
+                    q++;
+                if (nasm_isspace(*q)) {
+                    *q++ = '\0';
+                    while (*q && nasm_isspace(*q))
+                        q++;
+                }
+                if (!nasm_stricmp(v, "resident"))
+                    flags |= EXPDEF_FLAG_RESIDENT;
+                else if (!nasm_stricmp(v, "nodata"))
+                    flags |= EXPDEF_FLAG_NODATA;
+                else if (!nasm_strnicmp(v, "parm=", 5)) {
+                    bool err = false;
+                    flags |= EXPDEF_MASK_PARMCNT & readnum(v + 5, &err);
+                    if (err) {
+                        nasm_nonfatal("value `%s' for `parm' is non-numeric", v + 5);
+                        return DIRR_ERROR;
+                    }
+                } else {
+                    bool err = false;
+                    ordinal = readnum(v, &err);
+                    if (err) {
+                        nasm_nonfatal("unrecognised export qualifier `%s'", v);
+                        return DIRR_ERROR;
+                    }
+                    flags |= EXPDEF_FLAG_ORDINAL;
+                }
+            }
+
+            export = *exptail = nasm_malloc(sizeof(struct ExpDef));
+            exptail = &export->next;
+            export->next = NULL;
+            export->extname = nasm_strdup(extname);
+            export->intname = nasm_strdup(intname);
+            export->ordinal = ordinal;
+            export->flags = flags;
+        }
+        return DIRR_OK;
+
     default:
 	return DIRR_UNKNOWN;
     }
@@ -2505,7 +2499,7 @@ static void dbgbi_linnum(const char *lnfname, int32_t lineno, int32_t segto)
      */
     if (!any_segs) {
         int tempint = 0;
-        if (segto != obj_segment("__NASMDEFSEG", &tempint))
+        if (segto != obj_segment(DEFAULT_SEG, &tempint))
             nasm_panic("strange segment conditions in OBJ driver");
     }
 
@@ -2687,7 +2681,6 @@ const struct ofmt of_obj = {
     obj_stdmac,
     obj_init,
     null_reset,
-    nasm_do_legacy_output,
     obj_out,
     obj_deflabel,
     obj_segment,
